//	Author:		Dario Accornero
//	BeDev ID:	10183
//	Mailto:		adario@cs.bu.edu
//	Project:	BeScript
//	Module:
//	The BeScriptPeek window.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

//	System headers.

#ifndef _STDIO_H_
#include <stdio.h>
#endif
#ifndef _STDLIB_H_
#include <stdlib.h>
#endif
#ifndef	_STRING_H_
#include <string.h>
#endif
#ifndef _MENU_BAR_H
#include <MenuBar.h>
#endif
#ifndef _MENU_ITEM_H
#include <MenuItem.h>
#endif
#ifndef	_ALERT_H
#include <Alert.h>
#endif
#ifndef _SCREEN_H
#include <Screen.h>
#endif
#ifndef _PROPERTY_INFO_H
#include <PropertyInfo.h>
#endif

//	Local headers.

#include "Colors.h"
#include "ScriptingElement.h"
#include "ScriptingTargets.h"
#include "ScriptingUtils.h"
#include "BeScriptPeekApp.h"
#include "BeScriptPeekWin.h"
using namespace BeScriptPeek;

//	Local declarations.

namespace {
	const float infoViewHeight = B_H_SCROLL_BAR_HEIGHT + (B_H_SCROLL_BAR_HEIGHT / 2.0);
};

//	--- DOutlineListView:
//		Constructor.

DOutlineListView::DOutlineListView(	BRect frame,
									const char * name,
									list_view_type type,
									uint32 resizeMask,
									uint32 flags)
: BOutlineListView(frame, name, type, resizeMask, flags)
{
}

//	--- DOutlineListView:
//		Mouse pressure.

void DOutlineListView::MouseDown(BPoint where)
{
	//	Expanding/collapsing an item using B_OPTION_KEY means expanding/collapsing
	//	the whole item hierarchy.
	
	int32 index = FullListIndexOf(where);
	bool expanded = IsExpanded(index);
	BOutlineListView::MouseDown(where);
	if (index < 0 || index >= FullListCountItems())
		return;
	
	//	Only if we have changed the item expand state.
	
	bool state = IsExpanded(index);
	if (state != expanded)
	{
		BMessage* msg = Window()->CurrentMessage();
		int32 modifiers;
		if (msg->FindInt32("modifiers", &modifiers) != B_OK)
			return;
		if (modifiers & B_OPTION_KEY)
		{
			BListItem* item = FullListItemAt(index);
			if (item)
			{
				ExpandCollapseAction data(this, state);
				EachItemUnder(item, false, ExpandCollapse, (void*)&data);
			}
		}
	}
}

//	--- DOutlineListView:
//		Hierarchies expansion/collapse.

BListItem* DOutlineListView::ExpandCollapse(BListItem* item, void* data)
{
	ExpandCollapseAction* action = (ExpandCollapseAction*)data;
	if (action->state)
		action->list->Expand(item);
	else
		action->list->Collapse(item);
	return 0;
}

//	--- BeScriptPeekWindow:
//		Constructor.

BeScriptPeekWindow::BeScriptPeekWindow(const BRect& frame)
: BWindow(frame, "BeScriptPeek", B_DOCUMENT_WINDOW, B_ASYNCHRONOUS_CONTROLS)
{
	fCurrentTarget = -1;
	fMaxWidth = 0;
	fWatchLaunch = true;
	fWatchQuit = true;
	fDehydrateMode = BESCRIPTPEEK_FULL_DEHYDRATE;

	BRect winBounds(Bounds());
	CreateMenuBar(winBounds);	
	float menuBarHeight = fMenuBar->Frame().Height() + 1;
	ResizeTo(frame.Width(), frame.Height() + menuBarHeight);

	BRect aRect(Bounds());
	aRect.top = menuBarHeight;
	fBackView = new BView(aRect, "fBackView", B_FOLLOW_ALL, B_WILL_DRAW | B_FRAME_EVENTS);
	AddChild(fBackView);
	rgb_color backColor = ui_color(B_PANEL_BACKGROUND_COLOR);
	fBackView->SetViewColor(backColor);
	
	aRect.OffsetTo(0, 0);
	aRect.right /= 2;

	BRect bRect(aRect);
	bRect.OffsetBy(bRect.Width(), 0);
	bRect.right = fBackView->Bounds().right;

	bRect.right -= B_V_SCROLL_BAR_WIDTH;
	bRect.bottom -= B_H_SCROLL_BAR_HEIGHT;
	bRect.top += infoViewHeight;

	fView = new DOutlineListView(bRect,
								"fView",
								B_SINGLE_SELECTION_LIST,
								B_FOLLOW_ALL,
								B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE);
	fScroll = new BScrollView(	"fBackView->BScrollView",
								fView,
								B_FOLLOW_LEFT | B_FOLLOW_TOP,
								B_FRAME_EVENTS | B_WILL_DRAW,
								true,
								true);
	fBackView->AddChild(fScroll);
	fView->SetInvocationMessage(new BMessage(BESCRIPTPEEK_INVOKE_TARGET));
	fView->SetSelectionMessage(new BMessage(BESCRIPTPEEK_SELECT_TARGET));

	BRect iRect(bRect);
	iRect.bottom = iRect.top;
	iRect.top -= infoViewHeight;
	iRect.right += B_V_SCROLL_BAR_WIDTH;
	iRect.bottom -= 4;
	fSpecifier = new BStringView(iRect, "fSpecifier", "Specifier:");
	fBackView->AddChild(fSpecifier);

	aRect.InsetBy(4, 4);
	fBoxView = new BBox(aRect, "fBoxView", B_FOLLOW_LEFT | B_FOLLOW_TOP);
	fBoxView->SetLabel("Scripting Information");
	fBackView->AddChild(fBoxView);
	
	fListFrame = bRect;
	fTargetFrame = aRect;
	fListFrame.right += B_V_SCROLL_BAR_WIDTH;
	fListFrame.bottom += B_H_SCROLL_BAR_HEIGHT;

	BRect tRect(aRect);
	tRect.InsetBy(4, 4);
	tRect.top += 8;
	tRect.right -= B_V_SCROLL_BAR_WIDTH;
	tRect.bottom -= B_H_SCROLL_BAR_HEIGHT;
	tRect.right -= 8;
	tRect.bottom -= 8;
	fTargetView = new DOutlineListView(tRect,
								"fTargetView",
								B_SINGLE_SELECTION_LIST,
								B_FOLLOW_ALL,
								B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE);

	//	Locate suitable font. This looks okay in a not-too-small size.
	
	BFont font(be_fixed_font);
	font_family family;
	font_style style;
	if (FindFont("Monospac821 BT", "Roman", &family, &style))
	{
		font.SetFamilyAndStyle(family, style);
		font.SetSize(10);
	}
	else
		font.SetSize(9);
	fTargetView->SetFont(&font);

	fTargetScroll =	new BScrollView("fBoxView->BScrollView",
					fTargetView,
					B_FOLLOW_ALL,
					B_FRAME_EVENTS | B_WILL_DRAW,
					true,
					true);
	BScrollBar* hBar = fTargetScroll->ScrollBar(B_HORIZONTAL);
	BScrollBar* vBar = fTargetScroll->ScrollBar(B_VERTICAL);
	ScrollViewCorner* corner = new ScrollViewCorner(vBar->Frame().left, hBar->Frame().top);
	corner->SetViewColor(BeBackgroundGrey);
	fTargetScroll->AddChild(corner);
	fBoxView->AddChild(fTargetScroll);

	BScreen screen(this);
	SetSizeLimits(	defaultWindowWidth * 0.75,
					screen.Frame().Width(),
					defaultWindowHeight * 0.75,
					screen.Frame().Height());	
}

//	--- BeScriptPeekWindow:
//		Tell application to quit when closing.

bool BeScriptPeekWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

//	--- BeScriptPeekWindow:
//		Dispatch "add application" message and update window.

void BeScriptPeekWindow::MessageReceived(BMessage* msg)
{
	switch (msg->what)
	{
		case BESCRIPTPEEK_ADD_APPLICATION:
		{
			//	Add application hierarchy.
			
			ScriptingElement* objectToAdd;
			status_t result = msg->FindPointer(SF_RESULT, (void **)&objectToAdd);
			if (result == B_OK)
				if (fWatchLaunch)
					objectToAdd->AddToView(fView);
				else
					delete objectToAdd;
		}
		break;
		case BESCRIPTPEEK_REMOVE_APPLICATION:
		{
			//	Remove application hierarchy.

			ScriptingElement* objectToRemove;
			status_t result = msg->FindPointer(SF_RESULT, (void **)&objectToRemove);
			if (result == B_OK)
				if (fWatchQuit)
					fView->RemoveItem(objectToRemove);
		}
		break;
		case BESCRIPTPEEK_WATCH_LAUNCH:
		{
			fWatchLaunch = !fWatchLaunch;
			BMenuItem* item = fMenuBar->FindItem(BESCRIPTPEEK_WATCH_LAUNCH);
			if (item)
				item->SetMarked(fWatchLaunch);
		}
		break;
		case BESCRIPTPEEK_WATCH_QUIT:
		{
			fWatchQuit = !fWatchQuit;
			BMenuItem* item = fMenuBar->FindItem(BESCRIPTPEEK_WATCH_QUIT);
			if (item)
				item->SetMarked(fWatchQuit);
		}
		break;
		case BESCRIPTPEEK_GET_DYNAMIC:
		{
			ScriptingElement::DynamicNames = !ScriptingElement::DynamicNames;
			BMenuItem* item = fMenuBar->FindItem(BESCRIPTPEEK_GET_DYNAMIC);
			if (item)
				item->SetMarked(ScriptingElement::DynamicNames);
			UpdateIfNeeded();
			int32 items = fView->FullListCountItems();
			for (int32 item = 0; item < items; ++item)
			{
				ScriptingElement* element = (ScriptingElement*)fView->FullListItemAt(item);
				if (element)
				{
					element->ResetNames();
					if (ScriptingElement::DynamicNames)
						element->FindDynamicName();
					element->UpdateItem();
				}
			}
			fView->Invalidate();
		}
		break;
		case BESCRIPTPEEK_SYS_DEHYDRATE:
		case BESCRIPTPEEK_FULL_DEHYDRATE:
		case BESCRIPTPEEK_HIER_DEHYDRATE:
		{
			//	Make sure this is a valid selection.
				
			if (fDehydrateMode == msg->what)
				break;
			
			//	Change suites display mode: refresh scripting information.

			fDehydrateMode = msg->what;
			int32 curTarget = fCurrentTarget;

			fCurrentTarget = -1;
			BMessage selectMsg(BESCRIPTPEEK_SELECT_TARGET);
			selectMsg.AddInt32("index", curTarget);
			PostMessage(&selectMsg);
		}
		break;
		case BESCRIPTPEEK_INVOKE_TARGET:
			InvokeTarget(msg);
		break;
		case BESCRIPTPEEK_SELECT_TARGET:
			SelectTarget(msg);
		break;
		default:
			BWindow::MessageReceived(msg);
	}
}

//	--- BeScriptPeekWindow:
//		Adjust position of UI items on resize.

void BeScriptPeekWindow::FrameResized(float width, float height)
{
	BeginViewTransaction();
	BWindow::FrameResized(width, height);
	float half = (width / 2);
	BRect frame(fScroll->Frame());
	frame.bottom = height - B_H_SCROLL_BAR_HEIGHT - 4;
	frame.right = width + 2;
	frame.left = frame.right - half;
	fSpecifier->MoveTo(frame.left, fSpecifier->Frame().top);
	fSpecifier->ResizeTo(frame.Width(), fSpecifier->Frame().Height());
	fScroll->MoveTo(frame.left, frame.top);
	fScroll->ResizeTo(frame.Width(), frame.Height());
	fBoxView->ResizeTo((half - 8), (fBackView->Bounds().Height() - 8));
	UpdateScrollBars();
	EndViewTransaction();
}

//	--- BeScriptPeekWindow:
//		Create menu bar.

void BeScriptPeekWindow::CreateMenuBar(BRect frame)
{
	fMenuBar = new BMenuBar(frame, "BeScriptPeekWindow::MainMenu");

	BMenu* menu = new BMenu("Peek");
	fMenuBar->AddItem(menu);

	BMenuItem* item = new BMenuItem("Watch Application Launch", new BMessage(BESCRIPTPEEK_WATCH_LAUNCH), 'L');
	item->SetMarked(fWatchLaunch);
	menu->AddItem(item);
	item = new BMenuItem("Watch Application Quit", new BMessage(BESCRIPTPEEK_WATCH_QUIT), 'U');
	item->SetMarked(fWatchQuit);
	menu->AddItem(item);
	item = new BMenuItem("Retrieve Dynamic Names", new BMessage(BESCRIPTPEEK_GET_DYNAMIC), 'D');
	item->SetMarked(ScriptingElement::DynamicNames);
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("About BeScriptPeek"B_UTF8_ELLIPSIS, new BMessage(B_ABOUT_REQUESTED));
	item->SetTarget(be_app);
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Quit", new BMessage(B_QUIT_REQUESTED), 'Q');
	item->SetTarget(be_app);
	menu->AddItem(item);

	menu = new BMenu("Suites");
	menu->SetRadioMode(true);
	fMenuBar->AddItem(menu);
	item = new BMenuItem("System Display", new BMessage(BESCRIPTPEEK_SYS_DEHYDRATE));
	item->SetMarked(fDehydrateMode == BESCRIPTPEEK_SYS_DEHYDRATE);
	menu->AddItem(item);
	item = new BMenuItem("Expanded Display", new BMessage(BESCRIPTPEEK_FULL_DEHYDRATE));
	item->SetMarked(fDehydrateMode == BESCRIPTPEEK_FULL_DEHYDRATE);
	menu->AddItem(item);
	item = new BMenuItem("Hierarchical Display", new BMessage(BESCRIPTPEEK_HIER_DEHYDRATE));
	item->SetMarked(fDehydrateMode == BESCRIPTPEEK_HIER_DEHYDRATE);
	menu->AddItem(item);
	
	AddChild(fMenuBar);
}

//	--- BeScriptPeekWindow:
//		Select a target and retrieve its information.

void BeScriptPeekWindow::SelectTarget(BMessage* msg)
{
	//	Make sure we have a valid selection.

	int32 index;
	if (msg->FindInt32(SF_INDEX, &index) != B_OK)
		return;

	//	Only if we are not selecting the same thing again.
	
	if (index != fCurrentTarget)
	{		
		//	Destroy contents of target view (if any).
		
		DisableUpdates();
		fTargetView->MakeEmpty();
		EnableUpdates();

		//	Locate element. If it is invalid, we just quit.
		
		fCurrentTarget = index;
		ScriptingElement* element = (ScriptingElement*)fView->ItemAt(fCurrentTarget);
		if (!element)
		{
			//	Update window to reflect invalid selection.
			
			fSpecifier->SetText("Specifier:");
			fMaxWidth = 0;
			UpdateScrollBars();
			return;
		}

		//	Hierarchical display is handled by a different method.
		
		if (fDehydrateMode == BESCRIPTPEEK_HIER_DEHYDRATE)
			HierarchicalDehydrate(element);
		else
		{
			//	Iterate over suites and stuff them in a list.
			
			BMessage suites(element->GetSuites());
			BList list;
			int32 current = 0;
			fMaxWidth = 0;
			while (true)
			{
				//	Locate the suite name first.
				
				const char* suiteName;
				status_t error = suites.FindString(SF_SUITES, current, &suiteName);
				if (error != B_OK)
					break;
				
				//	Add description line.
				
				BString str("Supports ");
				str += suiteName;
				str += ":\\";
				char* text = new char[str.Length() + 1];
				sprintf(text, "%s", str.String());
				list.AddItem(text);
				
				//	Locate the suite properties and add them to the list.
				
				int32 sizefound;
				const void* voidptr;
				error = suites.FindData(SF_MESSAGES, B_PROPERTY_INFO_TYPE, current, &voidptr, &sizefound);
				if (error != B_OK)
					break;
	
				BPropertyInfo propinfo;
				if (propinfo.Unflatten(B_PROPERTY_INFO_TYPE, voidptr, sizefound) == B_OK)
				{
					if (fDehydrateMode == BESCRIPTPEEK_FULL_DEHYDRATE)
						FullDehydratePropertyInfo(&list, propinfo, sizefound);
					else if (fDehydrateMode == BESCRIPTPEEK_SYS_DEHYDRATE)
						DehydratePropertyInfo(&list, propinfo, sizefound);
				}
				
				//	Now, we create a suitable list to be displayed.
	
				DisableUpdates();
				int32 lines = list.CountItems();
				for (int32 line = 0; line < lines; ++line)
				{
					char* lineText = (char*)list.ItemAt(line);
					int len = 0;
					if (lineText && (len = strlen(lineText)) > 0)
					{
						bool disable = false;
						if (lineText[len - 1] == '\\')
						{
							disable = true;
							lineText[len - 1] = 0;
						}
						BStringItem* strItem = new BStringItem(lineText);
						if (disable)
							strItem->SetEnabled(false);
						fTargetView->AddItem(strItem);
						float width = fTargetView->StringWidth(lineText);
						if (width > fMaxWidth)
							fMaxWidth = width;
					}
					delete[] lineText;
				}
				list.MakeEmpty();
				EnableUpdates();
	
				//	Go over all available suites.
				
				++current;
			}
		}
		
		//	Adjust maximum width of target scroller.
		//	Also, adjust specifier string for human-readability.
		
		UpdateScrollBars();

		string str(specifier_to_str(element->GetAccessPtr()));
		BString textString("Specifier: ");
		textString += str.c_str();
		float width = fSpecifier->Frame().Width() - 4;
		fSpecifier->TruncateString(&textString, B_TRUNCATE_SMART, width);
		fSpecifier->SetText(textString.String());
	}
}

//	--- BeScriptPeekWindow:
//		Invoking a target equals to rescanning it if it's an application
//		or retrieving its dynamic name otherwise.

void BeScriptPeekWindow::InvokeTarget(BMessage* msg)
{
	//	Make sure we have a valid selection.

	int32 index;
	if (msg->FindInt32(SF_INDEX, &index) != B_OK)
		return;
	
	//	Extract the element. If it is an application, perform scan immediately.
	//	Note that we can use ItemAt() instead of FullListItemAt() because
	//	the item is guaranteed to be expanded when it was double-clicked.
	
	ScriptingElement* element = (ScriptingElement*)fView->ItemAt(index);
	if (element)
	{
		ScriptingApplication* app = dynamic_cast<ScriptingApplication*>(element);
		if (app)
		{
			app->Rescan(fView);
			fView->Invalidate();
		}
		else
		{
			//	Otherwise, just retrieve the dynamic name of the element,
			//	independently of the current dynamic name settings.

			element->ResetNames();
			element->FindDynamicName();
			element->UpdateItem();
			BRect frame = fView->ItemFrame(index);
			fView->Invalidate(frame);
		}		
	}
}

//	--- BeScriptPeekWindow:
//		Update the horizontal scrollbar for the scripting information box.
//		Currently the vertical scrollbar does not need updating.

void BeScriptPeekWindow::UpdateScrollBars()
{
	BScrollBar* hBar = fTargetScroll->ScrollBar(B_HORIZONTAL);

	float view_width = fTargetView->Frame().Width();
	float page_width = fMaxWidth + B_V_SCROLL_BAR_WIDTH;
	float width_prop = view_width / page_width;

	if (width_prop >= 1.0)
		hBar->SetRange(0.0, 0.0);
	else
	{
		float min = 0;
		float max = min + page_width - view_width;
		if (fTargetView->Bounds().left < min)
			min = fTargetView->Bounds().left;
		if (fTargetView->Bounds().left > max)
			max = fTargetView->Bounds().left;
		hBar->SetRange(min, max);
		hBar->SetSteps(ceil(view_width / 20), view_width);
	}

	hBar->SetProportion(width_prop);
}

//	--- BeScriptPeekWindow:
//		Create outline list containing hierarchical suites display.

void BeScriptPeekWindow::HierarchicalDehydrate(ScriptingElement* element)
{
	//	Create the outline list.
	
	BMessenger* messenger = element->GetMessenger();
	BMessage suites(element->GetSuites());
	BMessage access(element->GetAccess());
	InitRecursiveDehydration();
	int32 items = RecursiveDehydratePropertyInfo(messenger, &access, fTargetView, &suites);
	
	//	Compute widest row. Make sure we use the actual number of available rows.

	fMaxWidth = 0;
	items = fTargetView->FullListCountItems();
	for (int32 index = 0; index < items; ++index)
	{
		BStringItem* item = (BStringItem*)fTargetView->FullListItemAt(index);
		const char* lineText = item->Text();
		BRect frame(fTargetView->ItemFrame(index));
		float width = fTargetView->StringWidth(lineText);
		float widest = MAX(frame.right, width);
		if (widest > fMaxWidth)
			fMaxWidth = widest;
	}	
}
