//	Author:		Dario Accornero
//	BeDev ID:	10183
//	Mailto:		adario@cs.bu.edu
//	Project:	BeScript
//	Module:
//	The BeScriptPeek window.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

#ifndef	BESCRIPTPEEKWINDOW_H
#define	BESCRIPTPEEKWINDOW_H

#ifndef	_WINDOW_H
#include <Window.h>
#endif
#ifndef _OUTLINE_LIST_VIEW_H
#include <OutlineListView.h>
#endif
#ifndef	_SCROLL_VIEW_H
#include <ScrollView.h>
#endif
#ifndef _STRING_VIEW_H
#include <StringView.h>
#endif
#ifndef _BOX_H
#include <Box.h>
#endif

//	Local headers.

#include "ScrollViewCorner.h"

//	An augmented BOutlineListView that allows the expansion/collapse of
//	entire sub-hierarchies by clicking with the B_OPTION_KEY.

class DOutlineListView : public BOutlineListView
{
public:
	DOutlineListView(BRect frame,
					const char * name,
					list_view_type type = B_SINGLE_SELECTION_LIST,
					uint32 resizeMask = B_FOLLOW_LEFT | B_FOLLOW_TOP,
					uint32 flags = B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE);
	virtual	void 	MouseDown(BPoint where);
private:
	struct ExpandCollapseAction
	{
		ExpandCollapseAction(DOutlineListView* theList, bool theState)
		: list(theList), state(theState) {}
		DOutlineListView*	list;
		bool				state;
	};
	static BListItem* ExpandCollapse(BListItem* item, void* data);
};

//	The main window displays a hierarchy of scriptable objects.

class BeScriptPeekWindow : public BWindow
{
public:
	BeScriptPeekWindow(const BRect& frame);
	virtual bool QuitRequested();
	virtual void MessageReceived(BMessage* msg);
	virtual void FrameResized(float width, float height);
	
	inline DOutlineListView* ObjectsList() const	{ return fView; }
	inline bool WatchLaunch() const					{ return fWatchLaunch; }
	inline bool WatchQuit() const					{ return fWatchQuit; }
	
private:
	DOutlineListView*	fView;
	BScrollView*		fScroll;
	BMenuBar*			fMenuBar;
	BView*				fBackView;
	BBox*				fBoxView;
	BRect				fListFrame;
	BRect				fTargetFrame;
	DOutlineListView*	fTargetView;
	BScrollView*		fTargetScroll;
	BStringView*		fSpecifier;
	int32				fCurrentTarget;
	float				fMaxWidth;
	bool				fWatchLaunch;
	bool				fWatchQuit;
	uint32				fDehydrateMode;

	void CreateMenuBar(BRect frame);
	void SelectTarget(BMessage* msg);
	void InvokeTarget(BMessage* msg);
	void UpdateScrollBars();
	void HierarchicalDehydrate(ScriptingElement* element);
};

#endif	//	BESCRIPTPEEKWINDOW_H
