// Author:		Dario Accornero <adario@cs.bu.edu>
// BeDev ID:	10183
// Project:		BeScript
// Module:
// Scripting utilities.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

//	System headers.

#ifndef __CSTDIO__
#include <cstdio>
#endif
#ifndef __CSTDLIB__
#include <cstdlib>
#endif
#ifndef __CSTRING__
#include <cstring>
#endif
#ifndef _PATH_H
#include <Path.h>
#endif
#ifndef _ENTRY_H
#include <Entry.h>
#endif

//	Local headers.

#include "ScriptingElement.h"
#include "ScriptingUtils.h"

//	------------------------------------------------------
//	UTILITIES:
//	------------------------------------------------------

void print_message(BMessage* message)
{
	BList textlist;
	add_message_contents(&textlist, message, 0);
	
	printf("BMessage(%s):\n", get_datatype_string(message->what));
	for(int32 i=0;i<textlist.CountItems();i++){
		printf("   %s\n", (char*)textlist.ItemAt(i));
		free(textlist.ItemAt(i));
	}
}

void add_message_contents(BList* textlist, BMessage* msg, int32 level)
{
	int32 count;
	int32 i, sizefound, j;
	ulong typefound;
	char *namefound;
	void *voidptr;
	BMessage a_message;
	char *textline, *datatype, *content;
		
	// go though all message data
	count=msg->CountNames(B_ANY_TYPE);
	for(i=0;i<count;i++){
		msg->GetInfo(B_ANY_TYPE, i, &namefound, &typefound);
		j=0;
		
		while(msg->FindData(namefound, typefound, j++, (const void **)&voidptr, &sizefound)==B_OK){
			datatype=get_datatype_string(typefound);
			content=format_data(typefound, (char*)voidptr, sizefound);
			textline=(char*)malloc(20+level*4+strlen(namefound)+strlen(datatype)+strlen(content));
			memset(textline, 32, 20+level*4);
			sprintf(textline+level*4, "\"%s\" (%s) : %s", namefound, datatype, content);
			textlist->AddItem(textline);
			delete [] datatype;
			delete [] content;
			
			if(typefound==B_MESSAGE_TYPE){
				msg->FindMessage(namefound, j-1, &a_message);
				add_message_contents(textlist, &a_message, level+1);
			}else
			if(typefound==B_RAW_TYPE && strcmp(namefound, "_previous_")==0){
				if(a_message.Unflatten((const char *)voidptr)==B_OK){
					add_message_contents(textlist, &a_message, level+1);
				}
			}
		}		
	}
}

char* get_datatype_string(int32 type)
{
	char *str=new char[128];	
	switch(type){
		// data types
		case B_ANY_TYPE:	strcpy(str, "B_ANY_TYPE"); break;
		case B_ASCII_TYPE:	strcpy(str, "B_ASCII_TYPE"); break;
		case B_BOOL_TYPE:	strcpy(str, "B_BOOL_TYPE"); break;
		case B_CHAR_TYPE:	strcpy(str, "B_CHAR_TYPE"); break;
		case B_COLOR_8_BIT_TYPE:	strcpy(str, "B_COLOR_8_BIT_TYPE"); break;
		case B_DOUBLE_TYPE:	strcpy(str, "B_DOUBLE_TYPE"); break;
		case B_FLOAT_TYPE:	strcpy(str, "B_FLOAT_TYPE"); break;
		case B_GRAYSCALE_8_BIT_TYPE:	strcpy(str, "B_GRAYSCALE_8_BIT_TYPE"); break;
		case B_INT64_TYPE:	strcpy(str, "B_INT64_TYPE"); break;
		case B_INT32_TYPE:	strcpy(str, "B_INT32_TYPE"); break;
		case B_INT16_TYPE:	strcpy(str, "B_INT16_TYPE"); break;
		case B_INT8_TYPE:	strcpy(str, "B_INT8_TYPE"); break;
		case B_MESSAGE_TYPE:	strcpy(str, "B_MESSAGE_TYPE"); break;
		case B_MESSENGER_TYPE:	strcpy(str, "B_MESSENGER_TYPE"); break;
		case B_MIME_TYPE:	strcpy(str, "B_MIME_TYPE"); break;
		case B_MONOCHROME_1_BIT_TYPE:	strcpy(str, "B_MONOCHROME_1_BIT_TYPE"); break;
		case B_OBJECT_TYPE:	strcpy(str, "B_OBJECT_TYPE"); break;
		case B_OFF_T_TYPE:	strcpy(str, "B_OFF_T_TYPE"); break;
		case B_PATTERN_TYPE:	strcpy(str, "B_PATTERN_TYPE"); break;
		case B_POINTER_TYPE:	strcpy(str, "B_POINTER_TYPE"); break;
		case B_POINT_TYPE:	strcpy(str, "B_POINT_TYPE"); break;
		case B_RAW_TYPE:	strcpy(str, "B_RAW_TYPE"); break;
		case B_RECT_TYPE:	strcpy(str, "B_RECT_TYPE"); break;
		case B_REF_TYPE:	strcpy(str, "B_REF_TYPE"); break;
		case B_RGB_32_BIT_TYPE:	strcpy(str, "B_RGB_32_BIT_TYPE"); break;
		case B_RGB_COLOR_TYPE:	strcpy(str, "B_RGB_COLOR_TYPE"); break;
		case B_SIZE_T_TYPE:	strcpy(str, "B_SIZE_T_TYPE"); break;
		case B_SSIZE_T_TYPE	:	strcpy(str, "B_SSIZE_T_TYPE"); break;
		case B_STRING_TYPE:	strcpy(str, "B_STRING_TYPE"); break;
		case B_TIME_TYPE :	strcpy(str, "B_TIME_TYPE"); break;
		case B_UINT64_TYPE	:	strcpy(str, "B_UINT64_TYPE"); break;
		case B_UINT32_TYPE:	strcpy(str, "B_UINT32_TYPE"); break;
		case B_UINT16_TYPE :	strcpy(str, "B_UINT16_TYPE"); break;
		case B_UINT8_TYPE :	strcpy(str, "B_UINT8_TYPE"); break;
		case B_PROPERTY_INFO_TYPE: strcpy(str, "B_PROPERTY_INFO_TYPE"); break;
		// message constants:
		case B_ABOUT_REQUESTED :	strcpy(str, "B_ABOUT_REQUESTED"); break;
		case B_WINDOW_ACTIVATED :	strcpy(str, "B_WINDOW_ACTIVATED"); break;
		case B_ARGV_RECEIVED :	strcpy(str, "B_ARGV_RECEIVED"); break;
		case B_QUIT_REQUESTED :	strcpy(str, "B_QUIT_REQUESTED"); break;
		case B_CANCEL :	strcpy(str, "B_CANCEL"); break;
		case B_KEY_DOWN :	strcpy(str, "B_KEY_DOWN"); break;
		case B_KEY_UP :	strcpy(str, "B_KEY_UP"); break;
		case B_MINIMIZE	 :	strcpy(str, "B_MINIMIZE"); break;
		case B_MOUSE_DOWN :	strcpy(str, "B_MOUSE_DOWN"); break;
		case B_MOUSE_MOVED :	strcpy(str, "B_MOUSE_MOVED"); break;
		case B_MOUSE_ENTER_EXIT	 :	strcpy(str, "B_MOUSE_ENTER_EXIT"); break;
		case B_MOUSE_UP  :	strcpy(str, "B_MOUSE_UP"); break;
		case B_PULSE  :	strcpy(str, "B_PULSE"); break;
		case B_READY_TO_RUN :	strcpy(str, "B_READY_TO_RUN"); break;
		case B_REFS_RECEIVED :	strcpy(str, "B_REFS_RECEIVED"); break;
		case B_SCREEN_CHANGED :	strcpy(str, "B_SCREEN_CHANGED"); break;
		case B_VALUE_CHANGED :	strcpy(str, "B_VALUE_CHANGED"); break;
		case B_VIEW_MOVED :	strcpy(str, "B_VIEW_MOVED"); break;
		case B_VIEW_RESIZED :	strcpy(str, "B_VIEW_RESIZED"); break;
		case B_WINDOW_MOVED :	strcpy(str, "B_WINDOW_MOVED"); break;
		case B_WINDOW_RESIZED :	strcpy(str, "B_WINDOW_RESIZED"); break;
		case B_WORKSPACES_CHANGED :	strcpy(str, "B_WORKSPACES_CHANGED"); break;
		case B_WORKSPACE_ACTIVATED :	strcpy(str, "B_WORKSPACE_ACTIVATED"); break;
		case B_ZOOM	 :	strcpy(str, "B_ZOOM"); break;
		case _APP_MENU_ :	strcpy(str, "_APP_MENU_"); break;
		case _BROWSER_MENUS_ :	strcpy(str, "_BROWSER_MENUS_"); break;
		case _MENU_EVENT_  :	strcpy(str, "_MENU_EVENT_"); break;
		case _QUIT_  :	strcpy(str, "_QUIT_"); break;
		case _VOLUME_MOUNTED_  :	strcpy(str, "_VOLUME_MOUNTED_"); break;
		case _VOLUME_UNMOUNTED_	 :	strcpy(str, "_VOLUME_UNMOUNTED_"); break;
		case _MESSAGE_DROPPED_ 	 :	strcpy(str, "_MESSAGE_DROPPED_"); break;
		case _MENUS_DONE_ :	strcpy(str, "_MENUS_DONE_"); break;
		case _SHOW_DRAG_HANDLES_	 :	strcpy(str, "_SHOW_DRAG_HANDLES_"); break;
		case B_SET_PROPERTY		 :	strcpy(str, "B_SET_PROPERTY"); break;
		case B_GET_PROPERTY	 :	strcpy(str, "B_GET_PROPERTY"); break;
		case B_CREATE_PROPERTY	 :	strcpy(str, "B_CREATE_PROPERTY"); break;
		case B_DELETE_PROPERTY	 :	strcpy(str, "B_DELETE_PROPERTY"); break;
		case B_COUNT_PROPERTIES	 :	strcpy(str, "B_COUNT_PROPERTIES"); break;
		case B_EXECUTE_PROPERTY  :      strcpy(str, "B_EXECUTE_PROPERTY"); break;
		case B_GET_SUPPORTED_SUITES	 :	strcpy(str, "B_GET_SUPPORTED_SUITES"); break;
		case B_CUT 	 :	strcpy(str, "B_CUT"); break;
		case B_COPY 	 :	strcpy(str, "B_COPY"); break;
		case B_PASTE 	 :	strcpy(str, "B_PASTE"); break;
		case B_SELECT_ALL	 :	strcpy(str, "B_SELECT_ALL"); break;
		case B_SAVE_REQUESTED 	 :	strcpy(str, "B_SAVE_REQUESTED"); break;
		case B_MESSAGE_NOT_UNDERSTOOD :	strcpy(str, "B_MESSAGE_NOT_UNDERSTOOD"); break;
		case B_NO_REPLY 	 :	strcpy(str, "B_NO_REPLY"); break;
		case B_REPLY  :	strcpy(str, "B_REPLY"); break;
		case B_SIMPLE_DATA	 :	strcpy(str, "B_SIMPLE_DATA"); break;
		case B_ARCHIVED_OBJECT	 :	strcpy(str, "B_ARCHIVED_OBJECT"); break;
		case B_UPDATE_STATUS_BAR :	strcpy(str, "B_UPDATE_STATUS_BAR"); break;
		case B_RESET_STATUS_BAR	 :	strcpy(str, "B_RESET_STATUS_BAR"); break;
		case B_NODE_MONITOR		 :	strcpy(str, "B_NODE_MONITOR"); break;
		case B_QUERY_UPDATE	 :	strcpy(str, "B_QUERY_UPDATE"); break;
		case B_BAD_SCRIPT_SYNTAX: strcpy(str, "B_BAD_SCRIPT_SYNTAX"); break;

		// specifiers:
		case B_NO_SPECIFIER	 :	strcpy(str, "B_NO_SPECIFIER"); break;
		case B_DIRECT_SPECIFIER	 :	strcpy(str, "B_DIRECT_SPECIFIER"); break;
		case B_INDEX_SPECIFIER	 :	strcpy(str, "B_INDEX_SPECIFIER"); break;
		case B_REVERSE_INDEX_SPECIFIER	 :	strcpy(str, "B_REVERSE_INDEX_SPECIFIER"); break;
		case B_RANGE_SPECIFIER	 :	strcpy(str, "B_RANGE_SPECIFIER"); break;
		case B_REVERSE_RANGE_SPECIFIER	 :	strcpy(str, "B_REVERSE_RANGE_SPECIFIER"); break;
		case B_NAME_SPECIFIER	 :	strcpy(str, "B_NAME_SPECIFIER"); break;
		case B_ID_SPECIFIER	 :	strcpy(str, "B_ID_SPECIFIER"); break;

		case B_ERROR	 :	strcpy(str, "B_ERROR"); break;
				
		default:	// unknown
					id2str(type, str);
					break;
	}	
	return str;
}

char* format_data(int32 type, char* ptr, long size)
{
	char idtext[32];
	char *str;
	float *fptr;
	double *dptr;
	long i;
	entry_ref aref;
	BEntry entry;
	BPath path;
	int64 i64;
	int32 i32;
	int16 i16;
	int8 i8;
	uint64 ui64;
	uint32 ui32;
	uint16 ui16;
	uint8 ui8;
	BMessage anothermsg;
	BPropertyInfo propinfo;
	const property_info *pinfo;
	int32 pinfo_index;
	const value_info *vinfo;
	int32 vinfo_index, vinfo_count;
	char *tempstr;
		
	if(size<=0L){
		str=new char;
		*str=0;
		return str;
	}

	switch(type){
		case B_MIME_TYPE:
		case B_ASCII_TYPE:
		case B_STRING_TYPE:
					if(size>512) size=512;
					str=new char[size+4];
					*str='\"';
					strncpy(str+1, ptr, size);
					strcat(str, "\"");
					break;
		case B_POINTER_TYPE:
					str=new char[64];
					sprintf(str, "%p", *(void**)ptr);
					break;

		case B_REF_TYPE:
					str=new char[1024];
					anothermsg.AddData("myref", B_REF_TYPE, ptr, size);
					anothermsg.FindRef("myref", &aref);
					if(entry.SetTo(&aref)==B_OK){
						entry.GetPath(&path);
						strcpy(str, path.Path());
					}else{
						strcpy(str, "invalid entry_ref");
					}
					break;
					
		case B_SSIZE_T_TYPE:
		case B_INT64_TYPE:
					str=new char[64];
					i64=*(int64*)ptr;
					sprintf(str, "%Ld (0x%LX)", i64, i64);
					break;
		
		case B_SIZE_T_TYPE:
		case B_INT32_TYPE:
					str=new char[64];
					i32=*(int32*)ptr;
					sprintf(str, "%ld (0x%08lX)", i32, i32);
					break;
		
		case B_INT16_TYPE:
					str=new char[64];
					i16=*(int16*)ptr;
					sprintf(str, "%d (0x%04X)", i16, i16);
					break;
		
		case B_CHAR_TYPE:
		case B_INT8_TYPE:
					str=new char[64];
					i8=*(int8*)ptr;
					sprintf(str, "%d (0x%02X)", i8, i8);
					break;
		
		case B_UINT64_TYPE:
					str=new char[64];
					ui64=*(uint64*)ptr;
					sprintf(str, "%Lu (0x%LX)", ui64, ui64);
					break;
		
		case B_UINT32_TYPE:
					str=new char[64];
					ui32=*(uint32*)ptr;
					sprintf(str, "%lu (0x%08lX)", ui32, ui32);
					break;
		
		case B_UINT16_TYPE:
					str=new char[64];
					ui16=*(uint16*)ptr;
					sprintf(str, "%u (0x%04X)", ui16, ui16);
					break;
		
		case B_UINT8_TYPE:
					str=new char[64];
					ui8=*(uint8*)ptr;
					sprintf(str, "%u (0x%02X)", ui8, ui8);
					break;
		
		case B_BOOL_TYPE:
					str=new char[10];
					if(*ptr){
						strcpy(str, "TRUE");
					}else{
						strcpy(str, "FALSE");
					}
					break;
					
		case B_FLOAT_TYPE:
					str=new char[40];
					fptr=(float*)ptr;
					sprintf(str, "%.3f", *fptr);
					break;
					
		case B_DOUBLE_TYPE:
					str=new char[40];
					dptr=(double*)ptr;
					sprintf(str, "%.3f", *dptr);
					break;
					
		case B_RECT_TYPE:
					str=new char[200];
					fptr=(float*)ptr;
					sprintf(str, "BRect(%.1f, %.1f, %.1f, %.1f)", fptr[0], fptr[1], fptr[2], fptr[3]);
					break;
					
		case B_POINT_TYPE:
					str=new char[200];
					fptr=(float*)ptr;
					sprintf(str, "BPoint(%.1f, %.1f)", fptr[0], fptr[1]);
					break;

		case B_RGB_COLOR_TYPE:	
					str=new char[64];
					sprintf(str, "rgb_color(%u, %u, %u, %u)", ((uint8*)ptr)[0], ((uint8*)ptr)[1], ((uint8*)ptr)[2], ((uint8*)ptr)[3] );
					break;
					
		case B_COLOR_8_BIT_TYPE:
					str=new char[size*6+4];
					*str=0;
					for(i=0;i<min_c(256,size);i++){
						sprintf(idtext, "%u ", ((unsigned char*)ptr)[i]);
						strcat(str,idtext);
					}
					*(str+strlen(str)-2)=0;
					break;

		case B_MESSAGE_TYPE:
					str=new char[64];
					if(anothermsg.Unflatten((const char *)ptr)==B_OK){
						sprintf(str, "what=%s", get_datatype_string(anothermsg.what));	// LEAK
					}else{
						strcpy(str, "error when unflattening");
					}
					break;
					
		case B_PROPERTY_INFO_TYPE:
					if(propinfo.Unflatten(B_PROPERTY_INFO_TYPE, (const void *)ptr, size)==B_OK){
						str=new char[size*32];	// an approximation

						pinfo=propinfo.Properties();
						pinfo_index=0;

						sprintf(str, "\n        property   commands                            specifiers\n--------------------------------------------------------------------------------\n");
						
						while(pinfo_index<propinfo.CountProperties()){
						
							strcat(str,  "                "+(strlen(pinfo[pinfo_index].name) <16 ? strlen(pinfo[pinfo_index].name) : 16 ));
							strcat(str, pinfo[pinfo_index].name);
							strcat(str, "   ");
							char *start=str+strlen(str);

							for(int32 i=0;i<10 && pinfo[pinfo_index].commands[i];i++){
								tempstr=get_datatype_string(pinfo[pinfo_index].commands[i]);
								strcat(str, tempstr);
								strcat(str, " ");
								delete [] tempstr;
							}
							
							// pad the rest with spaces
							if(strlen(start)<36){
								strcat(str,  "                                    "+strlen(start) );
							}else{
								strcat(str,  "  " );
							}

							for(int32 i=0;i<10 && pinfo[pinfo_index].specifiers[i];i++){
								switch(pinfo[pinfo_index].specifiers[i]){
									case B_NO_SPECIFIER: strcat(str, "NONE "); break;
									case B_DIRECT_SPECIFIER: strcat(str, "DIRECT "); break;
									case B_INDEX_SPECIFIER: strcat(str, "INDEX "); break;
									case B_REVERSE_INDEX_SPECIFIER: strcat(str, "REV.INDEX "); break;
									case B_RANGE_SPECIFIER: strcat(str, "RANGE "); break;
									case B_REVERSE_RANGE_SPECIFIER: strcat(str, "REV.RANGE "); break;
									case B_NAME_SPECIFIER: strcat(str, "NAME "); break;
									case B_ID_SPECIFIER: strcat(str, "ID "); break;
									default:
									{
										char b[8];
										sprintf(str + strlen(str),
												"%s ",
												id2str(pinfo[pinfo_index].specifiers[i], b));
									}
									break;
								}
							}
							strcat(str, "\n");
							
							// is there usage info?
							if(pinfo[pinfo_index].usage){
								strcat(str, "                   Usage: ");
								strcat(str, pinfo[pinfo_index].usage);
								strcat(str, "\n");
							}						
							pinfo_index++;		// take next propertyinfo
						}
						
						// handle value infos....
						vinfo=propinfo.Values();
						vinfo_index=0;
						vinfo_count=propinfo.CountValues();

						if(vinfo && vinfo_count>0){
							sprintf(str+strlen(str), "\n            name   value                               kind\n--------------------------------------------------------------------------------\n");
							
							while(vinfo_index<vinfo_count){
							
								char *start=str+strlen(str);
								strcat(str,  "                "+(strlen(vinfo[vinfo_index].name) <16 ? strlen(vinfo[vinfo_index].name) : 16 ));
								strcat(str, vinfo[vinfo_index].name);
								strcat(str, "   ");
							
								sprintf(str+strlen(str), "0x%8lX (", vinfo[vinfo_index].value);
								id2str(vinfo[vinfo_index].value, str+strlen(str));
								strcat(str, ")");

								// pad the rest with spaces
								if(strlen(start)<36+19){
									strcat(str,  "                                                       "+strlen(start) );
								}else{
									strcat(str,  "  " );
								}

								switch(vinfo[vinfo_index].kind){
									case B_COMMAND_KIND: 	strcat(str, "COMMAND         "); break;
									case B_TYPE_CODE_KIND: 	strcat(str, "TYPE CODE       "); break;
									default:
									{
										char b[8];
										sprintf(str + strlen(str),
												"UNKNOWN!  %s",
												id2str(vinfo[vinfo_index].kind, b));
									}
									break;
								}								
	
								strcat(str, "\n");
								
								// is there usage info?
								if(vinfo[vinfo_index].usage){
									strcat(str, "                   Usage: ");
									strcat(str, vinfo[vinfo_index].usage);
									strcat(str, "\n");
								}
	
								vinfo_index++;		// take next valueinfo
							}
						}
						
					}else{
						str=new char[64];
						strcpy(str, "error when unflattening");
					}
					break;

		default:
					str=new char[min_c(256,size)*20+4];
					*str=0;
					for(i=0;i<min_c(256,size);i++){
						sprintf(idtext, "0x%02X, ", (uint16)ptr[i] );
						strcat(str,idtext);
					}
					*(str+strlen(str)-2)=0;
					break;
	}	
	return str;
}

char* id2str(int32 ID, char* here)
{
	uint8 digit0=(ID>>24)&255;
	uint8 digit1=(ID>>16)&255;
	uint8 digit2=(ID>>8)&255;
	uint8 digit3=(ID)&255;
	bool itsvalid=false;
	
	if(digit0==0){
		if(digit1==0){
			if(digit2==0){
				// 1 digits
				if(is_valid_char(digit3) ) itsvalid=TRUE;
				sprintf(here, "'%c'", digit3);
			}else{
				// 2 digits
				if(is_valid_char(digit2) && is_valid_char(digit3) ) itsvalid=TRUE;
				sprintf(here, "'%c%c'", digit2, digit3);
			}
		}else{
			// 3 digits
			if(is_valid_char(digit1) && is_valid_char(digit2) && is_valid_char(digit3) ) itsvalid=TRUE;
			sprintf(here, "'%c%c%c'", digit1, digit2, digit3);
		}
	}else{
		// 4 digits
		if(is_valid_char(digit0) && is_valid_char(digit1) && is_valid_char(digit2) && is_valid_char(digit3) ) itsvalid=TRUE;
		sprintf(here, "'%c%c%c%c'", digit0, digit1, digit2, digit3);
	}
	
	if(!itsvalid){
		sprintf(here, "%ldL", ID);
	}

	return here;
}

bool is_valid_char(uint8 c)
{
	return (c>=32 && c<128);
}

// Return a human-readable description of a specifier.

string specifier_to_str(const BMessage* msg, bool skipFirst)
{
	// First of all, make sure we have some specifiers in this message,
	// otherwise the result is empty.

	if (!msg->HasSpecifiers())
		return string("");

	// Now, check whether we have an autoresolving specifier.
	// Note that automatic specifiers work by default only using object names.
	// This is obvious, but since the scripting system is not particularly
	// well documented, we have to give all possible details.

	int32 autoresolve = -1;
	status_t error = msg->FindInt32("autoresolve", &autoresolve);
	if (error != B_OK)
		autoresolve = -1;

	// And now we can begin to build a string for this specifier.

	string str;
	int32 item = 0;
	BMessage specifier;
	while (msg->FindMessage(SF_SPECIFIERS, item++, &specifier) == B_OK)
	{
		bool isFirst = item == 1;
		char buf[256];
		if (!isFirst)
		{
			if (!skipFirst || (skipFirst && item > 2))
			{
				if (item == autoresolve + 1)
					str += " in ";
				else
					str += " of ";
			}
		}
		if (!isFirst || (isFirst && !skipFirst))
		{
			const char* property;
			specifier.FindString(SF_PROPERTY, &property);
			switch (specifier.what)
			{
				case B_NO_SPECIFIER:
					str += "(none)";
				break;
				case B_DIRECT_SPECIFIER:
					str += property;
				break;
				case B_INDEX_SPECIFIER:
				{
					str += property;
					int32 index;
					specifier.FindInt32(SF_INDEX, &index);
					sprintf(buf, " %ld", index);
					str += buf;
				}
				break;
				case B_REVERSE_INDEX_SPECIFIER:
				{
					str += property;
					int32 index;
					specifier.FindInt32(SF_INDEX, &index);
					sprintf(buf, " -%ld", index);
					str += buf;
				}
				break;
				case B_RANGE_SPECIFIER:
				{
					str += property;
					int32 index;
					specifier.FindInt32(SF_INDEX, &index);
					sprintf(buf, " [%ld", index);
					str += buf;
					int32 range;
					specifier.FindInt32(SF_RANGE, &range);
					sprintf(buf, " to %ld]", range);
					str += buf;
				}
				break;
				case B_REVERSE_RANGE_SPECIFIER:
				{
					str += property;
					int32 index;
					specifier.FindInt32(SF_INDEX, &index);
					sprintf(buf, " [-%ld", index);
					str += buf;
					int32 range;
					specifier.FindInt32(SF_RANGE, &range);
					sprintf(buf, " to %ld]", range);
					str += buf;
				}
				break;
				case B_NAME_SPECIFIER:
				{
					str += property;
					str += " \"";
					const char* name;
					specifier.FindString(SF_NAME, &name);
					str += name;
					str += "\"";
				}
				break;
				case B_ID_SPECIFIER:
				{
					str += property;
					str += " ";
					int32 id;
					specifier.FindInt32(SF_ID, &id);
					char idbuf[8];
					str += id2str(id, idbuf);
				}
				break;
				default:
				{
					char buf[8];
					str += property;
					str += " ";
					str += id2str(specifier.what, buf);
				}
			}
		}
	}
	return str;
}

// Return a human-readable description of a data message.

string data_to_str(const BMessage* msg)
{
	string str;
	int32 count = msg->CountNames(B_ANY_TYPE);
	for (int32 i = 0; i < count; ++i)
	{
		ulong typefound;
		char* namefound;
		msg->GetInfo(B_ANY_TYPE, i, &namefound, &typefound);
		
		int32 type = 0;
		void *voidptr;
		int32 sizefound;
		while (msg->FindData(	namefound,
								typefound,
								type++,
								(const void **)&voidptr,
								&sizefound) == B_OK)
		{
			if ((type + i) >= 2)
				str += " and ";
			char* content = format_data(typefound, (char*)voidptr, sizefound);
			
			// We add the data type name, but only if it's not "data".
			if (strcmp(namefound, SF_DATA))
			{
				str += namefound;
				str += "=";
			}
			str += content;
			delete[] content;
		}
	}
	return str;
}

//	Find the specified font.

bool FindFont(const char* fName, const char* fStyle, font_family* family, font_style* style)
{
	int32 numFamilies = count_font_families();
	font_family	localFamily;
	for (int32 i = 0; i < numFamilies; i++)
	{
		uint32 flags;
		if (get_font_family(i, &localFamily, &flags) == B_OK)
		{
			if (!strcmp((char*)localFamily, fName))
			{
				int32 numStyles = count_font_styles(localFamily);
				for (int32 j = 0; j < numStyles; j++)
				{
					font_style localStyle; 
					if (get_font_style(localFamily, j, &localStyle, &flags) == B_OK)
					{
						if (!strcmp((char*)localStyle, fStyle))
						{
							memcpy(family, &localFamily, sizeof(font_family));
							memcpy(style, &localStyle, sizeof(font_style));
							return true;
						}
					}
				}
			}
		}
	}	
	return false;
}

void DehydratePropertyInfo(BList* list, const BPropertyInfo& propinfo, int32 size, bool disable)
{
	//	Rows that must be disabled in the onscreen list are appended a blackslash on demand.
	//	They will be disabled and stripped of the blackslash by the onscreen list.
	
	const char* separator = "--------------------------------------------------------------------------------";
	const int stringSize = MAX(int32(strlen(separator) + 2), size * 8);
	const property_info* pinfo;
	int32 pinfo_index;
	int32 pinfo_count;
	char* tempstr;

	pinfo = propinfo.Properties();
	pinfo_index = 0;

	char* str = new char[stringSize];
	sprintf(str, "        property   commands                            specifiers%c", disable ? '\\' : 0);
	list->AddItem(str);
	str = new char[stringSize];
	sprintf(str, "%s%c", separator, disable ? '\\' : 0);
	list->AddItem(str);
	
	pinfo_count = propinfo.CountProperties();
	if (pinfo_count < 1)
		return;
	
	while (pinfo_index < pinfo_count)
	{
		str = new char[stringSize];
		*str = 0;
		strcat(str, "                " +
					(strlen(pinfo[pinfo_index].name) < 16 ? strlen(pinfo[pinfo_index].name) : 16));
		strcat(str, pinfo[pinfo_index].name);
		strcat(str, "   ");
		char* start = str + strlen(str);

		for (int32 i = 0; pinfo[pinfo_index].commands[i]; ++i)
		{
			tempstr = get_datatype_string(pinfo[pinfo_index].commands[i]);
			strcat(str, tempstr);
			strcat(str, " ");
			delete[] tempstr;
		}
		
		if (strlen(start) < 36)
			strcat(str,  "                                    " + strlen(start));
		else
			strcat(str,  "  ");

		for (int32 i = 0; pinfo[pinfo_index].specifiers[i]; ++i)
		{
			switch (pinfo[pinfo_index].specifiers[i])
			{
				case B_NO_SPECIFIER: strcat(str, "NONE "); break;
				case B_DIRECT_SPECIFIER: strcat(str, "DIRECT "); break;
				case B_INDEX_SPECIFIER: strcat(str, "INDEX "); break;
				case B_REVERSE_INDEX_SPECIFIER: strcat(str, "REV.INDEX "); break;
				case B_RANGE_SPECIFIER: strcat(str, "RANGE "); break;
				case B_REVERSE_RANGE_SPECIFIER: strcat(str, "REV.RANGE "); break;
				case B_NAME_SPECIFIER: strcat(str, "NAME "); break;
				case B_ID_SPECIFIER: strcat(str, "ID "); break;
				default: 
				{
					char b[8];
					sprintf(str + strlen(str),
							"%s ",
							id2str(pinfo[pinfo_index].specifiers[i], b));
				}
				break;
			}
		}
		list->AddItem(str);
		
		if (pinfo[pinfo_index].usage)
		{
			str = new char[stringSize];
			strcpy(str, "                   Usage: ");
			strcat(str, pinfo[pinfo_index].usage);
			if (disable)
				strcat(str, "\\");
			list->AddItem(str);
		}
			
		pinfo_index++;
	}
	
	const value_info* vinfo;
	int32 vinfo_index;
	int32 vinfo_count;
	
	vinfo = propinfo.Values();
	vinfo_index = 0;
	vinfo_count = propinfo.CountValues();

	if (vinfo && vinfo_count > 0)
	{
		str = new char[stringSize];
		sprintf(str, "%s%c", separator, disable ? '\\' : 0);
		list->AddItem(str);
		str = new char[stringSize];
		sprintf(str, "            name   value                               kind%c", disable ? '\\' : 0);
		list->AddItem(str);
		str = new char[stringSize];
		sprintf(str, "%s%c", separator, disable ? '\\' : 0);
		list->AddItem(str);
		
		while (vinfo_index < vinfo_count)
		{
			str = new char[stringSize];
			*str = 0;
			char *start = str;
			strcat(str, "                " + 
						(strlen(vinfo[vinfo_index].name) < 16 ? strlen(vinfo[vinfo_index].name) : 16));
			strcat(str, vinfo[vinfo_index].name);
			strcat(str, "   ");
		
			sprintf(str + strlen(str), "0x%8lX (", vinfo[vinfo_index].value);
			id2str(vinfo[vinfo_index].value, str + strlen(str));
			strcat(str, ")");

			if (strlen(start) < 36 + 19)
				strcat(str, "                                                       " + strlen(start));
			else
				strcat(str,  "  " );

			switch (vinfo[vinfo_index].kind)
			{
				case B_COMMAND_KIND: 	strcat(str, "COMMAND         "); break;
				case B_TYPE_CODE_KIND: 	strcat(str, "TYPE CODE       "); break;
				default:
				{
					char buf[8];
					strcat(str, "UNKNOWN!  ");
					strcat(str, id2str(vinfo[vinfo_index].kind, buf));
				}
				break;
			}
			list->AddItem(str);
			
			if (vinfo[vinfo_index].usage)
			{
				str = new char[stringSize];
				strcpy(str, "                   Usage: ");
				strcat(str, vinfo[vinfo_index].usage);
				if (disable)
					strcat(str, "\\");
				list->AddItem(str);
			}

			vinfo_index++;
		}
	}
	
	str = new char[1 + strlen(separator)];
	sprintf(str, "%s%c", separator, disable ? '\\' : 0);
	list->AddItem(str);
}

void FullDehydratePropertyInfo(BList* list, const BPropertyInfo& propinfo, int32 size, bool disable)
{
	//	Rows that must be disabled in the onscreen list are appended with a blackslash on demand.
	//	They will be disabled and stripped of the blackslash by the onscreen list.
	
	const char* separator = "--------------------------------------------------------------------------------";
	const int stringSize = MAX(int32(strlen(separator) + 2), size * 8);
	const property_info* pinfo;
	int32 pinfo_index;
	int32 pinfo_count;

	pinfo = propinfo.Properties();
	pinfo_index = 0;

	char* str = new char[stringSize];
	sprintf(str, "         command   property with specifier                          %c", disable ? '\\' : 0);
	list->AddItem(str);
	str = new char[stringSize];
	sprintf(str, "%s%c", separator, disable ? '\\' : 0);
	list->AddItem(str);
	
	pinfo_count = propinfo.CountProperties();
	if (pinfo_count < 1)
		return;
	
	while (pinfo_index < pinfo_count)
	{
		const char* property = pinfo[pinfo_index].name;
		int32 commandsFound = 0;
		for (int32 i = 0; pinfo[pinfo_index].commands[i]; ++i)
		{
			++commandsFound;
			char buf[256];
			switch (pinfo[pinfo_index].commands[i])
			{
				case B_SET_PROPERTY:		strcpy(buf, "set"); break;
				case B_GET_PROPERTY:		strcpy(buf, "get"); break;
				case B_CREATE_PROPERTY:		strcpy(buf, "create"); break;
				case B_DELETE_PROPERTY:		strcpy(buf, "delete"); break;
				case B_COUNT_PROPERTIES:	strcpy(buf, "count"); break;
				case B_EXECUTE_PROPERTY:    strcpy(buf, "execute"); break;
				case B_GET_SUPPORTED_SUITES:strcpy(buf, "getsuites"); break;
				case B_QUIT_REQUESTED:		strcpy(buf, "quit"); break;
				case B_ABOUT_REQUESTED:		strcpy(buf, "about"); break;
				case B_SAVE_REQUESTED:		strcpy(buf, "save"); break;
				case B_REFS_RECEIVED:		strcpy(buf, "load"); break;
				default:
				{
					char idstr[8];
					id2str(pinfo[pinfo_index].commands[i], idstr);
					sprintf(buf, "do %s", idstr);
				}
			}
			for (int32 i = 0; pinfo[pinfo_index].specifiers[i]; ++i)
			{
				str = new char[stringSize];
				*str = 0;
				sprintf(str, "%16s   ", buf);
				switch (pinfo[pinfo_index].specifiers[i])
				{
					case B_NO_SPECIFIER: sprintf(str + strlen(str), "%s (no specifier)", property); break;
					case B_DIRECT_SPECIFIER: strcat(str, property); break;
					case B_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s index (0,1,…)", property); break;
					case B_REVERSE_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s -index (0,1,…)", property); break;
					case B_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [index (0,1,…) to count (1,2,…)]", property); break;
					case B_REVERSE_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [-index (0,1,…) to count (1,2,…)]", property); break;
					case B_NAME_SPECIFIER: sprintf(str + strlen(str), "%s \"name\"", property); break;
					case B_ID_SPECIFIER: sprintf(str + strlen(str), "%s 'valu'", property); break;
					default:
					{
						char b[8];
						sprintf(str + strlen(str),
								"%s %s", property,
								id2str(pinfo[pinfo_index].specifiers[i], b));
					}
					break;
				}
				list->AddItem(str);
			}
		}
		if (!commandsFound)
		{
			int32 specifiersFound = 0;
			for (int32 i = 0; pinfo[pinfo_index].specifiers[i]; ++i)
			{
				++specifiersFound;
				str = new char[stringSize];
				*str = 0;
				sprintf(str, "%16s   ", "");
				switch (pinfo[pinfo_index].specifiers[i])
				{
					case B_NO_SPECIFIER: sprintf(str + strlen(str), "%s (no specifier)", property); break;
					case B_DIRECT_SPECIFIER: strcat(str, property); break;
					case B_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s index (0,1,…)", property); break;
					case B_REVERSE_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s -index (0,1,…)", property); break;
					case B_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [index (0,1,…) to count (1,2,…)]", property); break;
					case B_REVERSE_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [-index (0,1,…) to count (1,2,…)]", property); break;
					case B_NAME_SPECIFIER: sprintf(str + strlen(str), "%s \"name\"", property); break;
					case B_ID_SPECIFIER: sprintf(str + strlen(str), "%s 'valu'", property); break;
					default:
					{
						char b[8];
						sprintf(str + strlen(str),
								"%s %s", property,
								id2str(pinfo[pinfo_index].specifiers[i], b));
					}
					break;
				}
				list->AddItem(str);
			}
			if (!specifiersFound)
			{
				str = new char[stringSize];
				*str = 0;
				sprintf(str, "%16s   %s (any specifier)", "(any command)", property);
				list->AddItem(str);
			}
		}
		
		if (pinfo[pinfo_index].usage)
		{
			str = new char[stringSize];
			strcpy(str, "                   Usage: ");
			strcat(str, pinfo[pinfo_index].usage);
			if (disable)
				strcat(str, "\\");
			list->AddItem(str);
		}
			
		pinfo_index++;
	}
	
	const value_info* vinfo;
	int32 vinfo_index;
	int32 vinfo_count;
	
	vinfo = propinfo.Values();
	vinfo_index = 0;
	vinfo_count = propinfo.CountValues();

	if (vinfo && vinfo_count > 0)
	{
		str = new char[stringSize];
		sprintf(str, "%s%c", separator, disable ? '\\' : 0);
		list->AddItem(str);
		str = new char[stringSize];
		sprintf(str, "    command name   command value                       command kind%c", disable ? '\\' : 0);
		list->AddItem(str);
		str = new char[stringSize];
		sprintf(str, "%s%c", separator, disable ? '\\' : 0);
		list->AddItem(str);
		
		while (vinfo_index < vinfo_count)
		{
			str = new char[stringSize];
			*str = 0;
			char *start = str;
			strcat(str, "                " + 
						(strlen(vinfo[vinfo_index].name) < 16 ? strlen(vinfo[vinfo_index].name) : 16));
			strcat(str, vinfo[vinfo_index].name);
			strcat(str, "   ");
		
			char b[8];
			strcat(str, id2str(vinfo[vinfo_index].value, b));
			sprintf(str + strlen(str), " (0x%8lX)", vinfo[vinfo_index].value);

			if (strlen(start) < 36 + 19)
				strcat(str, "                                                       " + strlen(start));
			else
				strcat(str,  "  " );

			switch (vinfo[vinfo_index].kind)
			{
				case B_COMMAND_KIND: 	strcat(str, "COMMAND         "); break;
				case B_TYPE_CODE_KIND: 	strcat(str, "TYPE CODE       "); break;
				default:
				{
					char b[8];
					strcat(str, "UNKNOWN!  ");
					strcat(str, id2str(vinfo[vinfo_index].kind, b));
				}
				break;
			}
			list->AddItem(str);
			
			if (vinfo[vinfo_index].usage)
			{
				str = new char[stringSize];
				strcpy(str, "                   Usage: ");
				strcat(str, vinfo[vinfo_index].usage);
				if (disable)
					strcat(str, "\\");
				list->AddItem(str);
			}

			vinfo_index++;
		}
	}
	
	str = new char[1 + strlen(separator)];
	sprintf(str, "%s%c", separator, disable ? '\\' : 0);
	list->AddItem(str);
}

#ifndef BESCRIPT

namespace {
	BList	descentList;
};

void InitRecursiveDehydration()
{
	int32 items = descentList.CountItems();
	for (int32 index = 0; index < items; ++index)
	{
		BString* item = (BString*)descentList.ItemAt(index);
		delete item;
	}
	descentList.MakeEmpty();
}

int32 RecursiveDehydratePropertyInfo(BMessenger* msgr,
									BMessage* access,
									BOutlineListView* list,
									const BMessage* suites,
									int level)
{
	int32 linesAdded = 0;
	int32 current = 0;
	while (true)
	{
		//	Locate the suite name first.
		
		const char* suiteName;
		status_t error = suites->FindString(SF_SUITES, current, &suiteName);
		if (error != B_OK)
			break;
		
		//	Make sure we are not recurring over something already seen.
		
		bool dangerous = false;
		for (int32 i = 0; i < descentList.CountItems(); ++i)
		{
			BString* str = (BString*)descentList.ItemAt(i);
			if (*str == suiteName)
			{
				dangerous = true;
				break;
			}
		}
		if (dangerous)
		{
			BString header("Supports ");
			header += suiteName;
			header += " (displayed elsewhere)";
			BStringItem* item = new BStringItem(header.String(), level);
			item->SetEnabled(false);
			list->AddItem(item);
			++current;
			continue;
		}
		
		//	Add description line.
		
		BString header("Supports ");
		header += suiteName;
		header += ":";
		BStringItem* item = new BStringItem(header.String(), level);
		item->SetEnabled(false);
		list->AddItem(item);
		++linesAdded;
		
		//	Locate the suite properties and add them to the list.
		
		int32 sizefound;
		const void* voidptr;
		error = suites->FindData(SF_MESSAGES, B_PROPERTY_INFO_TYPE, current, &voidptr, &sizefound);
		if (error != B_OK)
			break;

		++current;
		BPropertyInfo propinfo;
		if (propinfo.Unflatten(B_PROPERTY_INFO_TYPE, voidptr, sizefound) != B_OK)
			continue;
		
		//	Perform property info expansion.
		
		const char* separator = "--------------------------------------------------------------------------------";
		const property_info* pinfo;
		int32 pinfo_index;
		int32 pinfo_count;
	
		pinfo = propinfo.Properties();
		pinfo_index = 0;
	
		item = new BStringItem("         command   property with specifier                          ", level);
		item->SetEnabled(false);
		list->AddItem(item);
		++linesAdded;
		item = new BStringItem(separator, level);
		item->SetEnabled(false);
		list->AddItem(item);
		++linesAdded;
		
		pinfo_count = propinfo.CountProperties();
		if (pinfo_count < 1)
			continue;
		
		while (pinfo_index < pinfo_count)
		{
			bool foundGET_SUITES = false;
			bool foundDIRECT = false;
			const char* property = pinfo[pinfo_index].name;
			int32 commandsFound = 0;
			for (int32 i = 0; pinfo[pinfo_index].commands[i]; ++i)
			{
				++commandsFound;
				char buf[256];
				switch (pinfo[pinfo_index].commands[i])
				{
					case B_SET_PROPERTY:		strcpy(buf, "set"); break;
					case B_GET_PROPERTY:		strcpy(buf, "get"); break;
					case B_CREATE_PROPERTY:		strcpy(buf, "create"); break;
					case B_DELETE_PROPERTY:		strcpy(buf, "delete"); break;
					case B_COUNT_PROPERTIES:	strcpy(buf, "count"); break;
					case B_EXECUTE_PROPERTY:    strcpy(buf, "execute"); break;
					case B_GET_SUPPORTED_SUITES:strcpy(buf, "getsuites"); foundGET_SUITES = true; break;
					case B_QUIT_REQUESTED:		strcpy(buf, "quit"); break;
					case B_ABOUT_REQUESTED:		strcpy(buf, "about"); break;
					case B_SAVE_REQUESTED:		strcpy(buf, "save"); break;
					case B_REFS_RECEIVED:		strcpy(buf, "load"); break;
					default:
					{
						char idstr[8];
						id2str(pinfo[pinfo_index].commands[i], idstr);
						sprintf(buf, "do %s", idstr);
					}
				}
				for (int32 i = 0; pinfo[pinfo_index].specifiers[i]; ++i)
				{
					char str[256];
					*str = 0;
					sprintf(str, "%16s   ", buf);
					switch (pinfo[pinfo_index].specifiers[i])
					{
						case B_NO_SPECIFIER: sprintf(str + strlen(str), "%s (no specifier)", property); break;
						case B_DIRECT_SPECIFIER: strcat(str, property); foundDIRECT = true; break;
						case B_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s index (0,1,…)", property); break;
						case B_REVERSE_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s -index (0,1,…)", property); break;
						case B_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [index (0,1,…) to count (1,2,…)]", property); break;
						case B_REVERSE_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [-index (0,1,…) to count (1,2,…)]", property); break;
						case B_NAME_SPECIFIER: sprintf(str + strlen(str), "%s \"name\"", property); break;
						case B_ID_SPECIFIER: sprintf(str + strlen(str), "%s 'valu'", property); break;
						default:
						{
							char b[8];
							sprintf(str + strlen(str),
									"%s %s", property,
									id2str(pinfo[pinfo_index].specifiers[i], b));
						}
						break;
					}
					item = new BStringItem(str, level);
					list->AddItem(item);
					++linesAdded;
				}
			}
			if (!commandsFound)
			{
				int32 specifiersFound = 0;
				for (int32 i = 0; pinfo[pinfo_index].specifiers[i]; ++i)
				{
					++specifiersFound;
					char str[256];
					*str = 0;
					sprintf(str, "%16s   ", "");
					switch (pinfo[pinfo_index].specifiers[i])
					{
						case B_NO_SPECIFIER: sprintf(str + strlen(str), "%s (no specifier)", property); break;
						case B_DIRECT_SPECIFIER: strcat(str, property); foundDIRECT = true; break;
						case B_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s index (0,1,…)", property); break;
						case B_REVERSE_INDEX_SPECIFIER: sprintf(str + strlen(str), "%s -index (0,1,…)", property); break;
						case B_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [index (0,1,…) to count (1,2,…)]", property); break;
						case B_REVERSE_RANGE_SPECIFIER: sprintf(str + strlen(str), "%s [-index (0,1,…) to count (1,2,…)]", property); break;
						case B_NAME_SPECIFIER: sprintf(str + strlen(str), "%s \"name\"", property); break;
						case B_ID_SPECIFIER: sprintf(str + strlen(str), "%s 'valu'", property); break;
						default:
						{
							char b[8];
							sprintf(str + strlen(str),
									"%s %s", property,
									id2str(pinfo[pinfo_index].specifiers[i], b));
						}
						break;
					}
					item = new BStringItem(str, level);
					list->AddItem(item);
					++linesAdded;
				}
				if (!specifiersFound)
				{
					char str[256];
					*str = 0;
					sprintf(str, "%16s   %s (any specifier)", "(any command)", property);
					item = new BStringItem(str, level);
					list->AddItem(item);
					++linesAdded;
				}
			}
			
			if (pinfo[pinfo_index].usage)
			{
				char str[256];
				strcpy(str, "                   Usage: ");
				strcat(str, pinfo[pinfo_index].usage);
				item = new BStringItem(str, level);
				item->SetEnabled(false);
				list->AddItem(item);
				++linesAdded;
			}
			
			if ((foundDIRECT && !commandsFound) || foundGET_SUITES)
			{
				BMessage curAccess(B_GET_SUPPORTED_SUITES);
				if (foundGET_SUITES)
				{
					//	Watch out: if the property supports GETSUITES, it might not support
					//	direct specifiers. Currently we ask for an index specifier.
					
					if (foundDIRECT)
						curAccess.AddSpecifier(property);
					else
						curAccess.AddSpecifier(property, int32(0));
				}
				else
					curAccess.AddSpecifier(property);
				int32 index = 0;
				BMessage specMsg;
				while (access->FindMessage(SF_SPECIFIERS, index++, &specMsg) == B_OK)
					curAccess.AddSpecifier(&specMsg);
				BMessage reply;

				error = msgr->SendMessage(&curAccess, &reply);
				if (error == B_OK)
				{
					if (!reply.IsEmpty())
					{						
						BString str("Scripting Information for \"");
						str += property;
						str += "\":";
						item = new BStringItem(str.String(), level, false);
						item->SetEnabled(false);
						list->AddItem(item);
						++linesAdded;

						BString* sN = new BString(suiteName);
						descentList.AddItem(sN);
						int32 lines = RecursiveDehydratePropertyInfo(msgr, &curAccess, list, &reply, level + 1);
						sN = (BString*)descentList.RemoveItem(descentList.CountItems() - 1);
						delete sN;

						if (lines > 0)
						{
							linesAdded += lines;
							str = "End of Scripting Information for \"";
							str += property;
							str += "\"";
							item = new BStringItem(str.String(), level);
							item->SetEnabled(false);
							list->AddItem(item);
							++linesAdded;
						}
						else
						{
							BString str(item->Text());
							str += " (none available)";
							item->SetText(str.String());
						}
					}
				}
			}
			
			pinfo_index++;
		}
		
		const value_info* vinfo;
		int32 vinfo_index;
		int32 vinfo_count;
		
		vinfo = propinfo.Values();
		vinfo_index = 0;
		vinfo_count = propinfo.CountValues();
	
		if (vinfo && vinfo_count > 0)
		{
			item = new BStringItem(separator, level);
			item->SetEnabled(false);
			list->AddItem(item);
			++linesAdded;
			item = new BStringItem("    command name   command value                       command kind", level);
			item->SetEnabled(false);
			list->AddItem(item);
			++linesAdded;
			item = new BStringItem(separator, level);
			item->SetEnabled(false);
			list->AddItem(item);
			++linesAdded;
			
			while (vinfo_index < vinfo_count)
			{
				char str[1024];
				*str = 0;
				char *start = str;
				strcat(str, "                " + 
							(strlen(vinfo[vinfo_index].name) < 16 ? strlen(vinfo[vinfo_index].name) : 16));
				strcat(str, vinfo[vinfo_index].name);
				strcat(str, "   ");
			
				char b[8];
				strcat(str, id2str(vinfo[vinfo_index].value, b));
				sprintf(str + strlen(str), " (0x%8lX)", vinfo[vinfo_index].value);
	
				if (strlen(start) < 36 + 19)
					strcat(str, "                                                       " + strlen(start));
				else
					strcat(str,  "  " );
	
				switch (vinfo[vinfo_index].kind)
				{
					case B_COMMAND_KIND: 	strcat(str, "COMMAND         "); break;
					case B_TYPE_CODE_KIND: 	strcat(str, "TYPE CODE       "); break;
					default:
					{
						char b[8];
						strcat(str, "UNKNOWN!  ");
						strcat(str, id2str(vinfo[vinfo_index].kind, b));
					}
					break;
				}
				item = new BStringItem(str, level);
				list->AddItem(item);
				++linesAdded;
				
				if (vinfo[vinfo_index].usage)
				{
					char str[256];
					strcpy(str, "                   Usage: ");
					strcat(str, vinfo[vinfo_index].usage);
					item = new BStringItem(str, level);
					item->SetEnabled(false);
					list->AddItem(item);
					++linesAdded;
				}
	
				vinfo_index++;
			}
		}
		
		item = new BStringItem(separator, level);
		item->SetEnabled(false);
		list->AddItem(item);
		++linesAdded;
	}
	
	return linesAdded;
}
#endif	// !BESCRIPT
