// Author:   Dario Accornero <adario@cs.bu.edu>
// BeDev ID: 10183
// Project:  BeScript

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

#ifndef YYSTYPE_H
#define YYSTYPE_H

// System headers.

#ifndef __STRING__
#include <string>
#endif
#ifndef __SGI_STL_LIST
#include <list>
#endif
#ifndef _SUPPORT_DEFS_H
#include <SupportDefs.h>
#endif
#ifndef _ENTRY_H
#include <Entry.h>
#endif
#ifndef _GRAPHICS_DEFS_H
#include <GraphicsDefs.h>
#endif

// Local headers.

#include "AbstractSyntax.h"
#include "ScriptingUtils.h"

// Semantic values for the BeScript grammar.
// This cannot be a union because it contains members with constructors.

typedef struct yystype
{
	// Message scripting contents:
	
	int8			int8val;
	int16			int16val;
	int32			int32val;
	float			floatval;
	double			doubleval;
	BPoint			pointval;
	BRect			rectval;
	rgb_color		colorval;
	entry_ref		fileval;

	int				ival;		// Integer literal.
	bool			bval;		// Boolean literal.
	double			fval;		// Floating point literal.
	string			sval;		// String literal or ID.
	uint32			wval;		// 'what' literal.

	Command*		cmd;		// Current command.
	CommandList*	block;		// Current command list block.
	DMessenger*		msgr;		// Current messenger.
	BMessage		spec;		// Current specifier.

	int				pos;		// Current token position.
}	YYSTYPE;

extern YYSTYPE yylval;

#endif	// YYSTYPE_H
